import casadi as ca
import matplotlib.pyplot as plt
import numpy as np

from rk4step import rk4step
from animate_pendulum import animate_pendulum

# for people who cannot see an interactive plot, uncomment the following lines
# import matplotlib
# if matplotlib.get_backend() == 'agg':
#     matplotlib.use('WebAgg')
# print(f'backend: {matplotlib.get_backend()}')



# continuous dynamics
nx = 2  # state dimension
nu = 1  # control dimension
N = 50  # horizon
dt = 0.1  # discretization time step
N_rk4 = 10  # number of rk4 steps per discretization time step
x0_bar = np.array([np.pi, 0])   # initial state

# build integrator
def dynamics(x, u):
    return ca.vertcat(x[1], ca.sin(x[0]) + u)
h = dt / N_rk4
x = ca.SX.sym('x', nx, 1)
u = ca.SX.sym('u', nu, 1)

# discrete dynamics
x_next = x
for _ in range(N_rk4):
    x_next = rk4step(x, u, dynamics, dt)

# integrator
F = ca.Function('F', [x, u], [x_next])

# cost
stage_cost = ca.Function('stage_cost', [x, u], [x.T @ x + 2 * u.T @ u])
terminal_cost = ca.Function('terminal_cost', [x], [10 * x.T @ x])

## NLP formulation
u_traj = ca.SX.sym('u_traj', N)               # vector of all controls
u_traj_init = .1 * np.ones((N, 1))         # intial guess U

# build vector of all states dependend on U (only needed for plotting)
# should have shape (ns, N+1)
x_traj = [x0_bar]

x_current = x0_bar
cost = 0

# note: you could use the casadi function fold here
for i in range(N):
    u_current = u_traj[i]
    cost += stage_cost(x_current, u_current)
    
    x_next = F(x_current, u_current)  
    
    x_current = x_next
    
    x_traj.append(x_next) # only needed for plotting

# terminal cost
cost += terminal_cost(x_current)

x_traj = ca.horzcat(*x_traj)

# plot sparsity pattern of the Hessian of the lagrangian
(hess, grad) = ca.hessian(cost, u_traj)
hess_cost = ca.Function('hess_cost', [u_traj], [hess])

plt.figure(1)
plt.spy(hess_cost(u_traj_init).full(), markersize=3)


# create nlp solver
nlp_seq = {'x': u_traj, 'f': cost}
solver_seq = ca.nlpsol('solver','ipopt', nlp_seq)

# solve nlp
sol_seq = solver_seq(x0=u_traj_init)

## visualize solution
u_traj_opt_seq = sol_seq['x'].full().T

F_u_traj = ca.Function('F_u_traj', [u_traj], [x_traj])
x_traj_opt_seq = F_u_traj(u_traj_opt_seq).full()

plt.figure(2)
plt.subplot(2,1,1)
plt.plot(x_traj_opt_seq.T)
plt.title('state trajectory')
plt.legend([r'$\theta$', r'$\omega$'])

plt.subplot(2,1,2)
plt.step(range(N), u_traj_opt_seq.T, where='post')
plt.title('control trajectory')
plt.legend([r'$\tau$'])
plt.xlabel('discrete time $k$')


ani = animate_pendulum(x_traj_opt_seq[0,:])

# ani.save('pendulum.gif')

plt.show()
